# -*- coding: utf-8 -*-
from collections import OrderedDict

from bunch import Bunch

from .columns import XlsxColumn, XlsxColumnFabric
from .core import XlsxField, logger


class XlsxRegion(XlsxField):
    PARENT_ATTR_NAME = 'worksheet'

    def __init__(self, relative_pos=(0, 0), **kwargs):
        self.relative_pos = relative_pos
        super(XlsxRegion, self).__init__(**kwargs)
        self.worksheet = None  # injected by worksheet

    def to_xlsx(self):
        self.before_to_xls()
        self._to_xlsx()
        self.after_to_xls()

    def before_to_xls(self):
        pass

    def after_to_xls(self):
        pass

    def _to_xlsx(self):
        raise NotImplementedError()


class XlsxTable(XlsxRegion):
    # columns here
    # clicks = IntegerXlsxColumn(u'Клики', width=27)
    # etc.

    table_style = None  # 'Table Style Medium 2'
    need_total = True
    total_row_format = None

    def __init__(self, **kwargs):
        super(XlsxTable, self).__init__(**kwargs)
        self.columns = OrderedDict()
        self.start_row = 0
        self.end_row = 0
        self._depended_columns = None
        self._undepended_columns = None
        self.extra_init()

    def extra_init(self):
        pass

    def remove_columns(self, *names):
        for name in names:
            try:
                del self.columns[name]
                delattr(self, name)
            except KeyError:
                logger.error("No column {} in {}".format(name, self.__class__.__name__))

    def _humanize_row(self, res_row):
        if not res_row:
            return
        for name, col in self.columns.iteritems():
            val = res_row[name]
            if val is None:
                res_row[name] = '--'
            else:
                res_row[name] = col.humanize(val)

    @property
    def depended_columns(self):
        if self._depended_columns is None:
            self._depended_columns = []
            for name, col in self.columns.iteritems():
                if col.is_depended:
                    self._depended_columns.append(col)
        return self._depended_columns

    @property
    def undepended_columns(self):
        if self._undepended_columns is None:
            self._undepended_columns = []
            for name, col in self.columns.iteritems():
                if not col.is_depended:
                    self._undepended_columns.append(col)
        return self._undepended_columns

    def get_data(self, humanize=False):
        self.before_get_data()
        res = []
        for col in self.columns.values():
            col._total = 0
        for db_row in self.get_queryset():
            if isinstance(db_row, dict):
                db_row = Bunch(db_row)
            res_row = Bunch()
            for col in self.undepended_columns:
                value = col.calc_value(source_row=db_row)
                res_row[col.name] = value
                if self.need_total:
                    col.calc_total(source_row=db_row, my_value=value, )
            for col in self.depended_columns:
                value = col.calc_value(source_row=db_row, res_row=res_row)
                res_row[col.name] = value
                if self.need_total:
                    col.calc_total(source_row=db_row, my_value=value, res_row=res_row)
            if humanize:
                self._humanize_row(res_row)
            res.append(res_row)

        if self.need_total and res:
            total_row = Bunch()
            for col in self.undepended_columns:
                total_row[col.name] = col.get_total()
            for col in self.depended_columns:
                total_row[col.name] = col.get_total(res_row=total_row)
            if humanize:
                self._humanize_row(total_row)
            res.append(total_row)

        return res

    def get_queryset(self):
        raise NotImplementedError()

    def _get_format(self, obj, format_attr, extra_fmt=None):
        cell_format = getattr(obj, format_attr, {}).copy()
        if extra_fmt:
            cell_format.update(extra_fmt)
        return self.worksheet.workbook.get_format(cell_format)

    def expand_columns(self):
        self.before_expand_columns()
        self.columns = OrderedDict()
        raw_columns = XlsxColumn.copy_fields_to_instance(instance=self)
        for name, column in raw_columns.iteritems():
            if isinstance(column, XlsxColumnFabric):
                for sub_column in column.get_columns():
                    sub_column.region = self
                    self.columns[sub_column.name] = sub_column
            else:
                self.columns[name] = column

    def before_expand_columns(self):
        pass

    def _to_xlsx(self):
        self.expand_columns()

        table_data = [[row[name] for name, col in self.columns.iteritems()] for row in self.get_data()]

        total_row = None
        if table_data and self.total_row_format:
            total_row = table_data[-1]
            table_data = table_data[:-1]
        options = {'data': table_data}
        if self.table_style:
            options.update({'style': self.table_style})
        columns = []
        for name, col in self.columns.iteritems():
            fmt = self._get_format(obj=col, format_attr='cell_format')
            columns.append(dict(header=col.title, format=fmt))
        options['columns'] = columns

        row_num, col_num = self.relative_pos
        self.worksheet.row_num += row_num

        self.start_row = self.worksheet.row_num
        self.worksheet.row_num += len(table_data)
        if self.total_row_format:
            self.worksheet.row_num += 1
        self.end_row = self.worksheet.row_num
        last_col = col_num + len(columns) - 1
        self.worksheet.out_ws.add_table(
            first_row=self.start_row,
            first_col=col_num,
            last_row=self.end_row,
            last_col=last_col,
            options=options
        )
        if total_row:
            for i, name in enumerate(self.columns):
                col = self.columns[name]
                fmt = self._get_format(obj=col, format_attr='cell_format', extra_fmt=self.total_row_format)
                self.worksheet.out_ws.write(self.worksheet.row_num, col_num + i, total_row[i], fmt)

    def set_column_width(self):
        for i, name in enumerate(self.columns):
            col = self.columns[name]
            self.worksheet.out_ws.set_column(i, i, col.width)

            # # старый код - м.б. пригодидзе для автоширины колонок
            # widths = [10 for col in table['header']]
            # for row in table['rows']:
            # for i, w in enumerate(widths):
            #         val = row[i]
            #         if isinstance(val, dict):
            #             continue
            #         cell_len = len(unicode(val))
            #         if w < cell_len:
            #             widths[i] = cell_len
            # for i, w in enumerate(widths):
            #     width = (w + 1) * 256
            #     if width > 65535:
            #         width = 65535
            #     sheet.col(i).width = width
            # sheet.set_panes_frozen(True)
            # sheet.set_horz_split_pos(1)
            # sheet.set_remove_splits(True)

    def before_get_data(self):
        pass


class XlsxUnrefinedTable(XlsxTable):
    head_format = {'bold': 1}
    total_row_format = None

    def _to_xlsx(self):
        self.expand_columns()

        rel_row_num, col_num = self.relative_pos
        self.worksheet.row_num += rel_row_num
        self.start_row = self.worksheet.row_num

        head_fmt = self._get_format(obj=self, format_attr='head_format')
        for i, name in enumerate(self.columns):
            col = self.columns[name]
            self.worksheet.out_ws.write(self.start_row, col_num + i, col.title, head_fmt)
        self.worksheet.row_num += 1

        data = self.get_data()
        if self.total_row_format:
            total_row = data[-1]
            data = data[:-1]
        else:
            total_row = []

        for row in data:
            for i, name in enumerate(self.columns):
                col = self.columns[name]
                fmt = self._get_format(obj=col, format_attr='cell_format')
                self.worksheet.out_ws.write(self.worksheet.row_num, col_num + i, row[col.name], fmt)
            self.worksheet.row_num += 1
        if self.total_row_format:
            for i, name in enumerate(self.columns):
                col = self.columns[name]
                fmt = self._get_format(obj=col, format_attr='cell_format', extra_fmt=self.total_row_format)
                self.worksheet.out_ws.write(self.worksheet.row_num, col_num + i, total_row[col.name], fmt)
            self.worksheet.row_num += 1
        self.end_row = self.worksheet.row_num


class XlsxChart(XlsxRegion):
    categories_column = 'period'
    main_type = 'line'
    main_series = [
        # Bunch(column='clicks', type='line', options={'y2_axis': True, 'line': {'color': '#ED7D31'}}),
    ]
    combain_type = None
    combain_series = None

    legend = None
    # legend = Bunch(position='bottom')
    size = None
    # size = Bunch(width=1557, height=405)
    chartarea = None
    # chartarea = Bunch(border={'none': True})
    title = {
        'name': u'Some name here',
        # 'name_font': {'size': 14, 'bold': False},
    }
    offset = None
    # offset = Bunch(x_offset25, y_offset=25)

    def __init__(self, source_region_name, **kwargs):
        super(XlsxChart, self).__init__(**kwargs)
        if not isinstance(source_region_name, str):
            raise Exception("Source must be name of region!")
        self.source_region_name = source_region_name
        self.source_region = None
        self.categories = None

    def _get_chart(self, chart_type, series):
        chart = self.worksheet.workbook.out_wb.add_chart({'type': chart_type})
        for dataset in series:
            serie_column = self.source_region.columns[dataset.column]
            options = dict(
                categories=self.categories,
                values=[self.worksheet.out_ws.name, self.source_region.start_row + 1, serie_column.position,
                        self.source_region.end_row, serie_column.position],
                name=serie_column.title,
            )
            if hasattr(dataset, 'options'):
                options.update(dataset.options)
            chart.add_series(options)
        return chart

    def _to_xlsx(self):
        try:
            self.source_region = self.worksheet.regions[self.source_region_name]
        except KeyError:
            raise Exception("Can't find region {} ".format(self.source_region_name))
        category_column = self.source_region.columns[self.categories_column]
        self.categories = [
            self.worksheet.out_ws.name,
            self.source_region.start_row + 1,
            category_column.position,
            self.source_region.end_row,
            category_column.position
        ]
        main_chart = self._get_chart(chart_type=self.main_type, series=self.main_series)
        if self.combain_series:
            combain_chart = self._get_chart(chart_type=self.combain_type, series=self.combain_series)
            main_chart.combine(combain_chart)

        if self.size:
            main_chart.set_size(self.size)
        if self.chartarea:
            main_chart.set_chartarea(self.chartarea)
        if self.legend:
            main_chart.set_legend(self.legend)
        if self.title:
            main_chart.set_title(self.title)

        chart_row = self.source_region.start_row + self.relative_pos[0]
        chart_col = category_column.position + self.relative_pos[1]
        offset = getattr(self, 'offset', {})
        self.worksheet.out_ws.insert_chart(chart_row, chart_col, main_chart, offset)




# -*- coding: utf-8 -*-

from .core import XlsxField
from .regions import XlsxRegion


class XlsxSheet(XlsxField):
    # regions here
    # top_table = XlsxTable()
    # pretty_table = OtherXlsxTable()
    # pretty_chart = XlsxChart()
    # etc.

    PARENT_ATTR_NAME = 'workbook'
    columns_width_region = None
    fill_priority = None

    def __init__(self, **kwargs):
        super(XlsxSheet, self).__init__(**kwargs)
        self.workbook = None  # injected by workbook
        self.regions = XlsxRegion.copy_fields_to_instance(instance=self)
        self.row_num = 0
        self.out_ws = None

    def get_name(self):
        return self.name

    def to_xlsx(self):
        self.row_num = 0
        self.prepare_to_xlsx()
        self.before_to_xlsx()
        self._context.update(**self.extra_context())
        for i, name in enumerate(self.regions):
            region = self.regions[name]
            region.to_xlsx()
            if name == self.columns_width_region or i == 0:
                region.set_column_width()
        self.after_to_xlsx()

    def prepare_to_xlsx(self):
        if not self.out_ws:
            sheet_name = self.get_name()
            self.out_ws = self.workbook.get_result_ws(sheet_name=sheet_name)

    @staticmethod
    def extra_context():
        return {}

    def before_to_xlsx(self):
        pass

    def after_to_xlsx(self):
        pass


class XlsxSheetFabric(XlsxSheet):
    def get_sheets(self):
        raise NotImplementedError()


class OneRegionXlsxSheet(XlsxSheet):
    region_class = None
    autoadded_region = None

    def __init__(self, **kwargs):
        if 'region_class' in kwargs:
            self.region_class = kwargs['region_class']
            del kwargs['region_class']
        elif not self.region_class:
            raise Exception('OneRegionXlsxSheet need region_class as attribute or in kwargs!')
        self.__class__.autoadded_region = self.region_class(**kwargs)
        super(OneRegionXlsxSheet, self).__init__(**kwargs)



import os

from setuptools import setup


README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='classy-xlsx',
    version='0.2.1',
    packages=['classy_xlsx'],
    license='BSD License',
    description='The package allows you to create xlsx files in style models Django ORM.',
    long_description=README,
    url='https://github.com/suguby/classy_xlsx',
    author='Shandrinov Vadim',
    author_email='suguby@gmail.com',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'XlsxWriter==0.7.2',
        'bunch==1.0.1',
    ]
)



# -*- coding: utf-8 -*-
from collections import OrderedDict
import logging
import copy

from bunch import Bunch


class XlsxContext(object):
    PARENT_ATTR_NAME = 'nope'

    def __init__(self, **kwargs):
        self._context = kwargs

    @property
    def parent(self):
        try:
            return getattr(self, self.PARENT_ATTR_NAME)
        except AttributeError:
            return None

    @property
    def context(self):
        context = Bunch(self.parent.context.copy()) if self.parent else Bunch()
        context.update(self._context)
        return context

    def update_context(self, **kwargs):
        self._context.update(kwargs)


class XlsxField(XlsxContext):
    __position = 0

    def __init__(self, **kwargs):
        super(XlsxField, self).__init__(**kwargs)
        self.name = None
        XlsxField.__position += 1
        self.position = XlsxField.__position

    @classmethod
    def copy_fields_to_instance(cls, instance, **attrs):
        fields = []
        for name in dir(instance.__class__):
            if name.startswith('_'):
                continue
            class_field = getattr(instance.__class__, name)
            if isinstance(class_field, cls):
                instance_field = copy.deepcopy(class_field)
                instance_field.name = name
                setattr(instance_field, cls.PARENT_ATTR_NAME, instance)
                for k, v in attrs.iteritems():
                    setattr(instance_field, k, v)
                setattr(instance, name, instance_field)
                fields.append(instance_field)
            if isinstance(class_field, tuple) and len(class_field) == 1 and isinstance(class_field[0], cls):
                print "Probably {} in {} has unnecessary comma!".format(name, instance.__name__)
        fields.sort(key=lambda x: x.position)
        for i, field in enumerate(fields):
            field.position = i
        return OrderedDict((field.name, field) for field in fields)


logger = logging.getLogger()




# -*- coding: utf-8 -*-
from bunch import Bunch

from .core import XlsxField


class XlsxColumn(XlsxField):
    _dependent = False
    PARENT_ATTR_NAME = 'region'
    quiet = False

    def __init__(self, title='', width=10, quiet=None, **kwargs):
        self.title = title
        self.width = width
        self.quiet = quiet if quiet is not None else self.__class__.quiet
        super(XlsxColumn, self).__init__(**kwargs)
        self.region = None  # injected by region
        self._kwargs = Bunch(kwargs)
        self._total = 0
        self.extra_init()

    def extra_init(self):
        pass

    @property
    def is_depended(self):
        return self._dependent

    def calc_value(self, source_row, res_row=None):
        val = source_row
        for name in self.name.split('__'):
            try:
                val = getattr(val, name, None)
                # if val.__class__.__name__ == 'RelatedManager':
                # val = val.first()
                # elif callable(val):
                #     val = val()
            except:
                if self.quiet:
                    return ''
                raise
        return val

    def calc_total(self, source_row, my_value, res_row=None):
        if my_value is not None:
            self._total += my_value

    def get_total(self, res_row=None):
        return self._total

    def __repr__(self):
        return self.name

    def humanize(self, value):
        return value

    def get_html_type(self):
        return 'text'


class FloatXlsxColumn(XlsxColumn):
    precision = 2
    cell_format = {'num_format': '# ##0.00'}

    def humanize(self, value):
        return round(float(value), self.precision)

    def get_html_type(self):
        return 'number'


class IntegerXlsxColumn(FloatXlsxColumn):
    precision = 0
    cell_format = {'num_format': '# ##0'}

    def humanize(self, value):
        return int(value)


class PercentXlsxColumn(FloatXlsxColumn):
    precision = 2
    cell_format = {'num_format': '0.00%'}

    def humanize(self, value):
        return round(float(value) * 100, self.precision)


class TextXlsxColumn(XlsxColumn):
    def calc_total(self, source_row, my_value, res_row=None):
        pass

    def get_total(self, res_row=None):
        return u''


class RatioColumn(XlsxColumn):
    dividend = 'redefine_me'  # делимое
    divisor = 'redefine_me'  # делитель
    _dependent = True

    def _get_value(self, res_row):
        if not res_row:
            return 0
        try:
            dividend = getattr(res_row, self.dividend)
            divisor = getattr(res_row, self.divisor)
        except AttributeError:
            return 0
        if divisor:
            return float(dividend) / float(divisor)
        return 0

    def calc_value(self, source_row, res_row=None):
        return self._get_value(res_row)

    def get_total(self, res_row=None):
        return self._get_value(res_row)


class WeightedAverageColumn(XlsxColumn):
    divisor = 'redefine_me'
    _dependent = True
    _total = 0  # PyCharm dummy

    def calc_total(self, source_row, my_value, res_row=None):
        if not res_row:
            return 0
        if my_value is None:
            return 0
        try:
            divisor = getattr(res_row, self.divisor)
        except AttributeError:
            return 0
        if divisor:
            self._total += float(my_value) * divisor

    def get_total(self, res_row=None):
        if not res_row:
            return 0
        try:
            divisor = getattr(res_row, self.divisor)
        except AttributeError:
            return 0
        if divisor:
            return self._total / float(divisor)
        return 0


class DateXlsxColumn(TextXlsxColumn):
    DATE_FORMAT = '%d.%m.%Y'

    def calc_value(self, source_row, res_row=None):
        value = super(DateXlsxColumn, self).calc_value(source_row, res_row)
        if value:
            return value.strftime(self.DATE_FORMAT)
        return value


class DateTimeXlsxColumn(DateXlsxColumn):
    DATE_FORMAT = '%d.%m.%Y %H:%M:%S'


class UnicodeXlsxColumn(TextXlsxColumn):
    def calc_value(self, source_row, res_row=None):
        value = super(UnicodeXlsxColumn, self).calc_value(source_row, res_row)
        return unicode(value)


class XlsxColumnFabric(XlsxColumn):
    def get_columns(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from collections import OrderedDict
from shutil import rmtree

from bunch import Bunch
import xlsxwriter

from .worksheet import XlsxSheetFabric, OneRegionXlsxSheet, XlsxSheet


class XlsxWorkbook(object):
    # sheets here
    # google = GoogleSheet()
    # yandex = YandexSheet()
    # etc.

    file_name = '/tmp/workbook.xlsx'

    def __init__(self, **kwargs):
        self._context = Bunch(kwargs)
        self.sheets = []
        self._dest_file = self.get_filename()
        self.out_wb = xlsxwriter.Workbook(self._dest_file)
        self.formats = {}
        self.tmp_dir = None

    @property
    def context(self):
        return self._context

    def get_format(self, fmt):
        key = repr(fmt)
        if key not in self.formats:
            self.formats[key] = self.out_wb.add_format(fmt)
        return self.formats[key]

    def get_filename(self):
        return self.file_name

    def get_result_ws(self, sheet_name):
        sheet_name = self._refine_sheet_name(sheet_name)
        for ws in self.out_wb.worksheets():
            ws_name = ws.get_name()
            if ws_name == sheet_name:
                break
        else:
            ws = self.out_wb.add_worksheet(sheet_name)
        return ws

    def make_report(self):
        self.sheets = OrderedDict()
        raw_sheets = XlsxSheet.copy_fields_to_instance(instance=self)
        for name, sheet in raw_sheets.iteritems():
            if isinstance(sheet, XlsxSheetFabric):
                i = 0
                for sub_sheet in sheet.get_sheets():
                    # TODO имя можно вынести в фабрику
                    sub_sheet_name = '{}_{}'.format(name, i)
                    sub_sheet.workbook = self
                    sub_sheet.name = sub_sheet_name
                    self.sheets[sub_sheet_name] = sub_sheet
                    i += 1
            else:
                self.sheets[name] = sheet

        self.before_make_report()
        sheets_by_priority = []
        for sheet in self.sheets.values():
            if sheet.fill_priority:
                sheet.prepare_to_xlsx()
                sheets_by_priority.append(sheet)
            else:
                sheet.to_xlsx()
        sheets_by_priority.sort(key=lambda x: x.fill_priority)
        for sheet in sheets_by_priority:
            sheet.to_xlsx()
        self.save()
        self.after_make_report()
        return self._dest_file

    def save(self):
        self.out_wb.close()
        if self.tmp_dir:
            rmtree(self.tmp_dir)
            self.tmp_dir = None

    def before_make_report(self):
        pass

    def after_make_report(self):
        pass

    def _refine_sheet_name(self, sheet_name):
        res = u''
        for char in sheet_name:
            if char in ('[', ']'):
                res += '_'
            else:
                res += char
        return res


class OneTableXlsxWorkbook(XlsxWorkbook):
    region_class = None
    list_name = 'list1'

    def __init__(self, **kwargs):
        if 'region_class' in kwargs:
            self.region_class = kwargs['region_class']
            del kwargs['region_class']
        elif not self.region_class:
            raise Exception('OneTableXlsxWorkbook need region_class as attribute or in kwargs!')
        sheet = OneRegionXlsxSheet(region_class=self.region_class)
        setattr(self.__class__, self.list_name, sheet)
        super(OneTableXlsxWorkbook, self).__init__(**kwargs)